#include <htc.h>
#include "common.h"
#include "lcd.h"
#include "tiny_printf.h"
#include "ad.h"

// See pic18f4550.h header
__CONFIG(1, CPUDIV1 & INTIO & FCMDIS);	// HSPLL, HS 
__CONFIG(2, PWRTDIS & BORDIS & WDTDIS);
__CONFIG(3, PBDIGITAL & MCLREN);
__CONFIG(4, LVPDIS & DEBUGEN & ICPORTDIS); // Successfully attach debugger thru PICKit3
//__CONFIG(4, DEBUGEN & XINSTDIS & LVPDIS & STVRDIS & ICPORTDIS);
__CONFIG(5, UNPROTECT);
__CONFIG(6, UNPROTECT);
__CONFIG(7, UNPROTECT);

void task_100ms();
void task_sw1();
void task_sw2();
void task_sw3_short();
void task_sw3_long();
void disp(unsigned char);

static long numint;
static char xtsk_100ms;
static char xtsk_10ms;

void interrupt main_isr(void)
{
	if(TMR0IF){
		// 1ms timer interrupt
		if(numint%10 == 0){
			// invoke 10ms task
			xtsk_10ms = 1;
			if(numint%100 == 0){
				// invoke 100ms task
				xtsk_100ms = 1;
			}
		}
		numint ++;
		TMR0L += (256 - 125);
		TMR0IF = 0;
	}
}

#define EN_INT()	(GIE=1)
#define DI_INT()	(GIE=0)

#define FAN_OFF() \
{ \
	RC1 = 0; \
	RA6 = 0; \
}

#define FAN_LO() \
{ \
	RC1 = 1; \
	RA6 = 0; \
}

#define FAN_HI() \
{ \
	RC1 = 1; \
	RA6 = 1; \
}

#define HEAT() \
{ \
	RC0 = 1; \
}

#define COOL() \
{ \
	RC0 = 0; \
}

void te_out(long duty)
{
	unsigned char temp;

	if(duty < 0){
		COOL();
		duty = -duty;
	} else if(duty > 0){ 
		HEAT();
	}

	if(duty > 1000){
		duty = 1000;
	}

	if(duty < 30){
		FAN_OFF();
	} else if(duty < 500){
		FAN_LO();
	} else {
		FAN_HI();
	}

	CCPR1L = (unsigned char)(duty >> 2);
	temp = CCP1CON;
	temp = temp & 0b11001111;
	temp |= ((unsigned char)(duty & 0x03)) << 5;
	CCP1CON = temp;
}


static double temp_obj = (double)23;
#define KP ((double)(0.10))
#define KI ((double)(0.10))
#define FILT_NUM1	(0.0305)
#define FILT_NUM2	(0.0305)
#define FILT_DEN2	(-0.9391)
#define SW_LONG 200
#define EEADR_TEMP_OBJ	0x00

void main(void)
{
	unsigned int sw1_cnt =0, sw2_cnt =0, sw3_cnt =0;

	// Oscillator
	OSCCON = 0b01110000;

	// Disable USB	
	USBEN = 0;
	UTRDIS = 1;

	// Setup port direction
	TRISA = 0b00001111;
	TRISB = 0b11100000;
	TRISC = 0x00;
	RBPU = 0;

	// Setup PWM
	CCP1CON = 0b00001100;
	PR2 = 0xff;
	CCPR1L = 0x10;
	T2CON = 0b00000101;

	//Setup Timer0
	T0CON = 0b11000011;
	TMR0IE = 1;
	EN_INT();

	lcd_init();
	ad_init();

	lcd_clear();
	FAN_OFF();
	COOL();

	// Load temp_obj
	temp_obj = eeprom_read(EEADR_TEMP_OBJ);
	if(temp_obj < 22 || temp_obj > 27){
		temp_obj = 22;
	}

	while(1){
		if(xtsk_10ms){
			// 10ms ^XN
			xtsk_10ms = 0;
			if(RB5 == 0){
				sw1_cnt++;
				if(sw1_cnt == 2){
					task_sw1();
				}
			} else {
				sw1_cnt = 0;
			}
			if(RB6 == 0){
				sw2_cnt++;
				if(sw2_cnt == 2){
					// SW2 Task
					task_sw2();
				}
			} else {
				sw2_cnt = 0;
			}
			if(RB7 == 0){
				sw3_cnt++;
				if(sw3_cnt == SW_LONG){
					// SW3 Task
					task_sw3_long();
				}
			} else {
				if(sw3_cnt && sw3_cnt < SW_LONG){
					task_sw3_short();
				}
				sw3_cnt = 0;
			}
		}

		if(xtsk_100ms){
			// 100ms^XN
			xtsk_100ms = 0;
			task_100ms();
		}
	}
}

double temp_in =0, temp_in_prev =0, temp_out =0, temp_out_prev =0;
double temp_err_prev =0, out = 0, out_prev =0;
double temp_min =99, temp_max = 0;
void task_100ms()
{
	double temp_raw, temp_err, delta;

	// Inside temp with LPF
	temp_raw = ad_gettemp(0);
	temp_in  = temp_raw * (double) FILT_NUM1 + temp_in_prev * (double)FILT_NUM2 - temp_in * (double)FILT_DEN2;
	temp_in_prev = temp_raw;

	// Outside temp with LPF
	temp_raw = ad_gettemp(1);
	temp_out = temp_raw * (double) FILT_NUM1 + temp_out_prev * (double)FILT_NUM2 - temp_out * (double)FILT_DEN2;
	temp_out_prev = temp_raw;

	// PI Control
	temp_err = temp_obj - temp_in;
	delta = KP * (temp_err - temp_err_prev) + KI * temp_err;
	out = out_prev + delta;
	temp_err_prev = temp_err;
	if(out > 1000){
		out = 1000;
	}
	if(out < -1000){
		out = -1000;
	}
	out_prev = out;
	te_out((long)out);
	disp(0);

	if(temp_in < temp_min){
		temp_min = temp_in;
	}
	if(temp_in > temp_max){
		temp_max = temp_in;
	}
}

void task_sw1()
{
	if(temp_obj < 27){
		temp_obj ++;
	} else {
		temp_obj = 22;
	}
	eeprom_write(EEADR_TEMP_OBJ, (unsigned char)temp_obj);
	disp(1);
}

void task_sw2()
{
	//disp(1);
	disp(2);
}

void task_sw3_short()
{
	disp(3);
}

void task_sw3_long()
{
	temp_min = temp_in;
	temp_max = temp_in;
	disp(4);
}

static char hold_count;
static char page;
void disp(unsigned char p)
{
	if(p){
		page = p;
		hold_count =30;
	}
	
	if(hold_count > 0){
		hold_count --;
	} else {
		page = 0;
	}

	switch(page){
	case 0:
		LCD_LINE(0);
		tiny_printf(lcd_putc, NULL, "%02ld.%ld  %02ld", 
			(unsigned long)temp_in, (unsigned long) (temp_in*10) %10, (unsigned long)temp_obj);
		LCD_LINE(1);
		{	
			long u10s, u1s;
			char pol;
			pol = out < 0 ? '-' : '+';
			u10s = (long)out / 10;
			u1s = (long)out % 10;			
			if(u1s < 0) u1s = -u1s;
			if(u10s < 0) u10s = -u10s;
			tiny_printf(lcd_putc, NULL, "%c%3ld.%ld\\%  ", pol, u10s, u1s);
		}
		break;
	case 1:
		LCD_LINE(0);
		tiny_printf(lcd_putc, NULL, "Temp set");
		LCD_LINE(1);
		tiny_printf(lcd_putc, NULL, "%ld C     ", (long)temp_obj);
		break;
	case 2:
		LCD_LINE(0);
		tiny_printf(lcd_putc, NULL, "Out temp");
		LCD_LINE(1);
		tiny_printf(lcd_putc, NULL, "%02ld.%ld     ", 
			(unsigned long)temp_out, (unsigned long) (temp_out*10) %10);
		break;
	case 3:
		LCD_LINE(0);
		tiny_printf(lcd_putc, NULL, "MIN %02ld.%ld", (unsigned long)temp_min, (unsigned long) (temp_min*10)%10);
		LCD_LINE(1);
		tiny_printf(lcd_putc, NULL, "MAX %02ld.%ld", (unsigned long)temp_max, (unsigned long) (temp_max*10)%10);
		break;
	case 4:
		LCD_LINE(0);
		tiny_printf(lcd_putc, NULL, "MIN/MAX ");
		LCD_LINE(1);
		tiny_printf(lcd_putc, NULL, "CLEAR   ");
		break;		

	default:
		break;
	}
}